.caption {
    text-align: center;
    font-size:15px;
    }

#img {width: 100%}

/* sidebar */
.sidebar {
    padding-left: 20px;
    margin-top: 20px;
    margin-bottom: 20px;
}

/* all links */
.sidebar .nav>li>a {
    color: #999;
    border-left: 2px solid transparent;
    padding: 4px 20px;
    font-size: 13px;
    font-weight: 400;
}

/* nested links */
.sidebar .nav .nav>li>a {
    padding-top: 1px;
    padding-bottom: 1px;
    padding-left: 30px;
    font-size: 12px;
}

/* active & hover links */
.sidebar .nav>.active>a, 
.sidebar .nav>li>a:hover, 
.sidebar .nav>li>a:focus {
    color: #563d7c;                 
    text-decoration: none;          
    background-color: transparent;  
    border-left-color: #563d7c; 
}
/* all active links */
.sidebar .nav>.active>a, 
.sidebar .nav>.active:hover>a,
.sidebar .nav>.active:focus>a {
    font-weight: 700;
}
/* nested active links */
.sidebar .nav .nav>.active>a, 
.sidebar .nav .nav>.active:hover>a,
.sidebar .nav .nav>.active:focus>a {
    font-weight: 500;
}

/* nested active links */
.sidebar .nav .nav .nav>li>a{
    padding-left: 50px;
}

/* hide inactive nested list */
.sidebar .nav ul.nav {
    display: none;           
}
/* show active nested list */
.sidebar .nav>.active>ul.nav {
    display: block;           
}

/* don't be affected by <section> indentation */
.sidebar * {
    margin-left: 0px;
}
