body{
    margin: 0 auto;
    font-family: Georgia, Palatino, serif;
    color: #444444;
    line-height: 1;
    max-width: 960px;
    padding: 30px;
}
h1, h2, h3, h4, h5, h6 {
    color: #111111;
    font-weight: 400;
}
h1, h2, h3, h4, h5, h6, p {
    margin-bottom: 24px;
    padding: 0;
}
h1 {
    font-size: 48px;
}
h2 {
    font-size: 40px;
}
h3 {
    font-size: 32px;
}
h4, h5, h6 {
    font-size: 28px;
}
a {
    color: #0099ff;
    margin: 0;
    padding: 0;
    vertical-align: baseline;
}
a:hover {
    text-decoration: none;
    color: #ff6600;
}
a:visited {
    color: purple;
}
ul, ol {
    padding: 0;
    margin-left: 48px;
}
li {
    line-height: 24px;
}
li ul, li ul {
    margin-left: 24px;
}
p, ul, ol {
    font-size: 16px;
    line-height: 24px;
    max-width: 540px;
}
pre {
    padding: 0px 24px;
    max-width: 800px;
    white-space: pre-wrap;
}
code {
    font-family: Consolas, Monaco, Andale Mono, monospace;
    line-height: 1.5;
    font-size: 13px;
}
aside {
    display: block;
    float: right;
    width: 390px;
}
blockquote {
    border-left:.5em solid #eee;
    padding: 0 2em;
    margin-left:0;
    max-width: 476px;
}
blockquote  cite {
    font-size:14px;
    line-height:20px;
    color:#bfbfbf;
}
blockquote cite:before {
    content: '\2014 \00A0';
}

blockquote p {
    color: #666;
    max-width: 460px;
}

hr {
    width: 540px;
    text-align: left;
    margin: 0 auto 0 0;
    color: #999;
}

table, th, td {
    border:1px solid #DDDDDD;
    border-collapse:collapse;
    font-weight:normal
}

th {
    background-color: #E6E6E6;
}

tr.alt td {
    background-color: #F8F8F8;
}

:target {
    background-color: #F5F9FC;
    padding-top: 10px;
    padding-bottom: 15px;
    padding-left: 10px;
}
